/*
 * File:   HashT.h
 * Author: peter
 *
 * Created on 03. Dezember 2012, 13:54
 */

#include <math.h>
#include <cstdlib>
#include <time.h>
#include "Position.h"
#include "globals.h"

#ifndef HASHT_H
#define	HASHT_H

class HashT {
public:
    HashT() {};
    
    /***********************************************************
     * Initialisiere Transposition-Hash-Table                  *
     * - generiere Zufallszahlen für Zobrist-Hash              *
     * - initialisiere Hash-Map entsprechend angegebener Größe *
     * - setze alle Werte in der Hash-Map auf 0                *
     * - setze aktuelle transpositionAge auf 0                 *
     ***********************************************************/
    void init(unsigned int size) {
        this->size=size;
        hashMask=(size-1)>>2;
        ageMask=18302628885633695744UL; // die höchsten 7 bits
        depthMask=144110790029344768UL; // die nächsten 15 bits
        typeMask=3298534883328UL; // die nächsten 2 bits
        bestmoveMask=1095216660480UL; // die nächsten 8 bits
        scoreMask=4294967295UL; // die niedrigsten 32 bits
        
        ageOffset=57;
        depthOffset=42;
        typeOffset=40;
        bestmoveOffset=32;        
        
        map = new unsigned long[size];
        clearHash();
        
        // srand(time(NULL));
        srand(3133742);
        for (int i=0;i<64;i++) {
            kb[i]=((unsigned long)rand()<<32)|rand();
            qb[i]=((unsigned long)rand()<<32)|rand();
            bb[i]=((unsigned long)rand()<<32)|rand();
            nb[i]=((unsigned long)rand()<<32)|rand();
            rb[i]=((unsigned long)rand()<<32)|rand();
            pb[i]=((unsigned long)rand()<<32)|rand();
            kw[i]=((unsigned long)rand()<<32)|rand();
            qw[i]=((unsigned long)rand()<<32)|rand();
            bw[i]=((unsigned long)rand()<<32)|rand();
            nw[i]=((unsigned long)rand()<<32)|rand();
            rw[i]=((unsigned long)rand()<<32)|rand();
            pw[i]=((unsigned long)rand()<<32)|rand();            
        }
        side2move=((unsigned long)rand()<<32)|rand();
        for (unsigned int i=0;i<16;i++)
            castling[i]=((unsigned long)rand()<<32)|rand();
        for (unsigned int i=0;i<8;i++)
            files[i]=((unsigned long)rand()<<32)|rand();
    };
    
    /*****************************************
     * Setze alle Werte der Hah-Map auf 0    *
     * Setze aktuelle transpositionAge auf 0 *
     *****************************************/
    void clearHash() {
        transpositionAge=0;
        for (unsigned int i=0;i<size;i++)
            map[i]=0UL;
    };
    
    /**********************************************************
     * Berechne einen Zobrist-Hash basierend auf              *
     * der aktuellen Position und den bei der Initialisierung *
     * gesetzen Zufallszahlen.                                *
     **********************************************************/
    unsigned long getHashKey(Position *pos) {     
        unsigned long fullHash=0L;
        for (int i=0;i<64;i++) {
            if (figOnPos((*pos).kings[0],i))       fullHash^=kw[i];
            if (figOnPos((*pos).queens[0],i))      fullHash^=qw[i];
            if (figOnPos((*pos).bishops[0],i))     fullHash^=bw[i];
            if (figOnPos((*pos).knights[0],i))     fullHash^=nw[i];
            if (figOnPos((*pos).rooks[0],i))       fullHash^=rw[i];
            if (figOnPos((*pos).pawns[0],i))       fullHash^=pw[i];
            if (figOnPos((*pos).kings[1],i))       fullHash^=kb[i];
            if (figOnPos((*pos).queens[1],i))      fullHash^=qb[i];
            if (figOnPos((*pos).bishops[1],i))     fullHash^=bb[i];
            if (figOnPos((*pos).knights[1],i))     fullHash^=nb[i];
            if (figOnPos((*pos).rooks[1],i))       fullHash^=rb[i];
            if (figOnPos((*pos).pawns[1],i))       fullHash^=pb[i];                        
        }

        if (pos->onMove)
            fullHash^=side2move;
        
        short int cast=0;
        cast^=pos->castlingRules;
        fullHash^=castling[cast];
        
        fullHash^=files[(pos->ep)%8];
        
        return fullHash;
    };
    
    /**************************************************************
     * Speichert einen Hash-Key mit dem entsprechenden Score      *
     * und zusätzlichen Informationen in der Transposition-Table. *
     **************************************************************/
    void insert(unsigned long hashKey, int score, unsigned int nodeType, unsigned int bestmove, unsigned int currentDepth) {

        /*******************************************************
         * Die aktuelle transpositionAge, die aktuelle Tiefe,  *
         * der Typ der aktuellen Position und der Score werden *
         * zu einem 64bit-Hash-Value zusammengefasst, der ggf. *
         * sofern dieselbe Position nochmals erreicht wird,    *
         * abgefragt werden kann.                              *
         *******************************************************/
        unsigned int shortenHash=(unsigned int)hashKey&hashMask;
        unsigned long newHashValue=((unsigned long)transpositionAge)<<ageOffset;
        newHashValue|=((unsigned long)currentDepth)<<depthOffset;
        newHashValue|=((unsigned long)nodeType)<<typeOffset;
        newHashValue|=((unsigned long)bestmove)<<bestmoveOffset;
        newHashValue|=(unsigned)score;
        
        /**************************************************************************
         * Das einfügen des HashKeys bzw. des HashValues kann auf drei            *
         * verschiedene Arten erfolgen:                                           *
         * - der HashKey (dh. die Position) ist bereits in der Transpositon-Table *
         *   gespeichert:                                                         *
         *   sofern die gespeicherte Tiefe der Position kleiner als die aktuelle  *
         *   Tiefe ist, kann der alte (schlechtere) Score durch den neuen         *
         *   (besseren) Score ersetzt werden.                                     *
         * - ist der HaskKey nicht gespeichert, wird nach Slots gesucht,          *
         *   deren transpositionAge älter als die aktuelle ist, dh. der Score     *
         *   stammt aus einer früheren Suche. Gibt es solche Slots, wird          *
         *   derjenige Slot überschrieben, beim die gespeicherte Tiefe am         *
         *   niedrigsten ist.                                                     *
         * - gibt es auch keinen Slot aus einer früheren Suche, dann wird         *
         *   der Slot mit der niedrigsten Tiefe überschrieben.                    *
         **************************************************************************/
        short int replaceIndex=-1;
        unsigned short int storedDepth=9999;
        
        for (unsigned short int i=0;i<4;i++) {
            unsigned long storedKey=map[shortenHash+i*2];
            unsigned long storedValue=map[shortenHash+i*2+1];

            /**********************************************************
             * HashKey bereits gespeichert, also Score überschreiben. *
             **********************************************************/
            if (storedKey==hashKey) {
                if (((storedValue&depthMask)>>depthOffset)<=currentDepth) {
                    map[shortenHash+i*2+1]=newHashValue;
                    return;
                }
            }
            
            /****************************************************
             * nach Slots aus früheren Suchvorgängen suchen und *
             * den mit der niedrigsten Tiefe auswählen.         *
             ****************************************************/
            unsigned int storedAge=(unsigned int)((map[shortenHash+i*2+1]&ageMask)>>ageOffset);
            if (storedAge != transpositionAge   &&      ((storedValue&depthMask)>>depthOffset)<storedDepth       ) {
                replaceIndex=i;
                storedDepth=((storedValue&depthMask)>>depthOffset);
            }
            
            /******************************************************************
             * den Slot mit der niedrigsten Tiefe auswählen, sofern           *
             * noch keine HashKeys aus früheren Suchvorgängen gefunden wurden *
             ******************************************************************/
            if ( replaceIndex==-1 &&       ((storedValue&depthMask)>>depthOffset)<storedDepth   ) {
                replaceIndex=i;
                storedDepth=((storedValue&depthMask)>>depthOffset);
            }
            
        }

        /********************************************************
         * HashKey und HashValue an ermitteltem Index speichern *
         ********************************************************/
        map[shortenHash+replaceIndex*2]=hashKey;
        map[shortenHash+replaceIndex*2+1]=newHashValue;
    };
    
    /************************************************************************
     * Prüft ob eine Position in der Trannsposition-Table gespeichert ist   *
     * und ob wir den gespeicherten Wert auch tatsächlich verwenden können. *
     ************************************************************************/
    bool contains(unsigned long hashKey, unsigned int currentDepth, int *score, int alpha, int beta, int *bestmoveHash) {
        
        /****************************************************************************
         * Ein Zobrist Hash ist 64bit lang, dh. es gibt 2^64 verschiedene Hashes.   *
         * Da wir aber nicht 2^64 Werte speichern können und wir den Hash als       *
         * Array-Index (für einen schnellen Zugriff) verwenden möchten, wir der     *
         * 64bit-Hash mit der hashMask auf eine akzeptable Länge verkürzt.          *
         * Dadurch kann einerseits direkt auf einen bestimmen Array-Index           *
         * zugegriffen werden, allerdings entstehen aber Kollisionen mit dem        *
         * verkürzten Hash. Deswegen gibt es für jeden Array-Index 4 Slots.         *
         * In jedem Slot wird der gesamte 64bit-Hash mit dem entsprechen 64bit-Wert *
         * gespeichert. Beim Zugriff auf einen bestimmten 64bit-Hash müssen daher   *
         * alle 4 Slots überprüft werden um festzustellen, ob es sich bei einen von *
         * ihnen um den gesuchten Zobrist-Hash handelt.                             *
         */
        unsigned int shortenHash=(unsigned int)hashKey&hashMask;
        
        for (unsigned short int i=0;i<4;i++) {
            unsigned long storedKey=map[shortenHash+i*2];
            unsigned long hashDepth=(map[shortenHash+i*2+1]&depthMask)>>depthOffset;
            
            /***************************
             * Position wurde gefunden *
             ***************************/
            if (storedKey==hashKey) {

                /********************************************************
                 * Den Index des besten Zuges aus dem Hash extrahieren. *
                 ********************************************************/
                (*bestmoveHash)=(int)(map[shortenHash+i*2+1]&bestmoveMask);
                
                /*******************************************************************
                 * unabhängig ob wir den gespeicherten Score verwenden können,     *
                 * wird die transpositionAge auf den aktuellen Wert aktualisiert.  *
                 * Dadurch wird sichergestellt, dass sie nicht überschrieben wird. *
                 *******************************************************************/
                map[shortenHash+i*2+1]=(map[shortenHash+i*2+1]&(~ageMask))|((unsigned long)transpositionAge<<ageOffset);
                
                /**************************************************************
                 * Der gespeicherte Score kann nur dann verwendet werden,     *
                 * wenn die Position bis zu einer Tiefe evaluiert wurde, die  *
                 * zumindest so hoch ist, als wenn die Position von der       *
                 * aktuellen Tiefe evaluiert werden würde.                    *
                 * Außerdem wird überprüft, welchem Node-Typ die gespeicherte *
                 * Position angehört (exakt, beta, alpha).                    *
                 **************************************************************/
                if (hashDepth>=currentDepth) {
                    
                    unsigned int nodetype=(map[shortenHash+i*2+1]&typeMask)>>typeOffset;
                    
                    int scr=(int)(map[shortenHash+i*2+1]&scoreMask);
                    
                    if (nodetype==1) { // exakt
                        (*score)=scr;
                        return true;
                    }
                    if (nodetype==2 && scr>=beta) { // beta
                        (*score) = beta;
                        return true;
                    }
                    if (nodetype==3 && scr<=alpha) { // alpha
                        (*score) = alpha;
                        return true;
                    }
                }
                break;
            }
        }
        /***********************************************
         * Hash der gesuchten Position nicht gefunden. *
         ***********************************************/
        return false;
    };
    
    /********************************
     * Zeiger auf Map-Array löschen *
     ********************************/
    virtual ~HashT() { delete[] map; };
    
    /**********************************************
     * addiert '1' zur aktuellen transpositionAge *
     **********************************************/
    void inkTranspositionAge() { transpositionAge++; }

private:
    unsigned int size;
    unsigned int transpositionAge;
    unsigned long hashMask;
    unsigned long depthMask;
    unsigned long ageMask;
    unsigned long typeMask;
    unsigned long bestmoveMask;
    unsigned long scoreMask;
    
    unsigned short int depthOffset;
    unsigned short int ageOffset;
    unsigned short int typeOffset;
    unsigned short int bestmoveOffset;
    
    unsigned long kb[64];
    unsigned long qb[64];
    unsigned long bb[64];
    unsigned long nb[64];
    unsigned long rb[64];
    unsigned long pb[64];
    unsigned long kw[64];
    unsigned long qw[64];
    unsigned long bw[64];
    unsigned long nw[64];
    unsigned long rw[64];
    unsigned long pw[64];         
    unsigned long side2move;
    unsigned long castling[16];
    unsigned long files[8];
    
    unsigned long *map;
};

#endif	/* HASHT_H */