/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   UCTNodePlayoutSimilarityA.cpp
 * Author: peter
 * 
 * Created on 10. April 2014, 07:45
 */

#include "UCTNodePlayoutSimilarityA.h"

UCTNodePlayoutSimilarityA::UCTNodePlayoutSimilarityA() {
    numChilds=0;
    totalValue=0;
    visits=0;
    parent=NULL;
}

UCTNodePlayoutSimilarityA::UCTNodePlayoutSimilarityA(string lastMoveString, unsigned long lastMoveFrom, unsigned long lastMoveTo, int lastMovePromotion, UCTNodePlayoutSimilarityA *parent) {
    numChilds=0; // from normal constructor
    totalValue=0;
    visits=0;
    this->parent=parent; // additional
    this->lastMoveString=lastMoveString;
    this->lastMoveFrom=lastMoveFrom;
    this->lastMoveTo=lastMoveTo;
    this->lastMovePromotion=lastMovePromotion;
}

UCTNodePlayoutSimilarityA::~UCTNodePlayoutSimilarityA() {
    for (int i=0;i<numChilds;i++) {
        if (childs[i]!=NULL) {
            delete childs[i];
            childs[i]=NULL;
        }
    }
}

string UCTNodePlayoutSimilarityA::getMove() {
    if (numChilds==0)
        return "0000";
    int mostVisitedValue=childs[0]->visits;
    int mostVisitedIndex=0;
    for (int i=1;i<numChilds;i++) {
        if (childs[i]->visits>mostVisitedValue) {
            mostVisitedValue=childs[i]->visits;
            mostVisitedIndex=i;
        }
    }
    return childs[mostVisitedIndex]->lastMoveString;
}

UCTNodePlayoutSimilarityA* UCTNodePlayoutSimilarityA::select(PositionMonteCarloTreeSearch *rootPos) {
    PositionMonteCarloTreeSearch *currentPos=rootPos;
    bool blackToMove=currentPos->isBlackToMove();
    UCTNodePlayoutSimilarityA *cur=this;
    UCTNodePlayoutSimilarityA *chosen=this;
    double bestVal=-1;
    // wähle solange child-knoten mittels uct-methode bis
    // knoten entweder noch keine child-knoten hat oder
    // noch nicht alle childs des knoten erstellt wurden
    while (cur->hasChildren()) {
        if ((cur->isNotFullyExpanded()))
            return cur;
        chosen=cur->getChild(0);
        for (int i=0;i<cur->numChilds;i++) {
            double winningRate=(double)(cur->getChild(i)->totalValue)/(cur->getChild(i)->visits);
            if (blackToMove)
                winningRate=1-winningRate;
            double uctVal=winningRate+sqrt(2*log(cur->visits)/cur->getChild(i)->visits);
            if (uctVal>=bestVal) {
                bestVal=uctVal;
                chosen=cur->getChild(i);
            }
        }
        currentPos->move_figure(chosen->lastMoveFrom,chosen->lastMoveTo,chosen->lastMovePromotion);
        blackToMove=!blackToMove;
        cur=chosen;
    }
    return cur;
}

UCTNodePlayoutSimilarityA* UCTNodePlayoutSimilarityA::expand(PositionMonteCarloTreeSearch *rootPos) {
    // knoten hat noch überhaupt keine child-knoten oder
    // es wurden noch nicht alle child-knoten erstellt
    // wenn knoten eine schach-matt position ist,
    // braucht der knoten auch nicht weiter expandiert werden.
    PositionMonteCarloTreeSearch *currentPos=rootPos; // new PositionMonteCarloTreeSearch(*rootPos);
    if (currentPos->isMate())
        return this;
    string moveArray[256];
    int moveIndex[256];
    unsigned long moveFrom[256];
    unsigned long moveTo[256];
    int movePromotion[256];
    cntMoves=currentPos->getMCTSChilds(moveArray,moveIndex,moveFrom,moveTo,movePromotion);
    if (numChilds==cntMoves)
        return this;
    string lastMoveString = moveArray[numChilds];
    unsigned long lastMoveFrom = moveFrom[numChilds];
    unsigned long lastMoveTo = moveTo[numChilds];
    int lastMovePromotion = movePromotion[numChilds];
    return createAndAddChild(lastMoveString, lastMoveFrom, lastMoveTo, lastMovePromotion);
}

double UCTNodePlayoutSimilarityA::simulate(PositionMonteCarloTreeSearch *rootPos) {
    PositionMonteCarloTreeSearch *currentPos=rootPos;
    int cntMoves=0;
    while(1) {
        currentPos->doRandomMoveSimilarityA();
        cntMoves++;
        if  (cntMoves>80)
            break;
        else if (currentPos->isKingWhiteLost())
            return 0;
        else if (currentPos->isKingBlackLost())
            return 1;
    }
    return 0.5;
}

void UCTNodePlayoutSimilarityA::update(double value) {
    totalValue+=value;
    visits++;
    if (parent!=NULL)
        parent->update(value);
}

bool UCTNodePlayoutSimilarityA::hasChildren() {
    return numChilds>0;
}

bool UCTNodePlayoutSimilarityA::isNotFullyExpanded() {
    return (cntMoves!=numChilds);
}

UCTNodePlayoutSimilarityA* UCTNodePlayoutSimilarityA::getChild(int i) {
    return (childs[i]!=NULL)?childs[i]:NULL;
}

UCTNodePlayoutSimilarityA* UCTNodePlayoutSimilarityA::createAndAddChild(string lastMoveString, unsigned long lastMoveFrom, unsigned long lastMoveTo, int lastMovePromotion) {
    childs[numChilds]= new UCTNodePlayoutSimilarityA(lastMoveString, lastMoveFrom, lastMoveTo, lastMovePromotion,this);
    numChilds++;
    return childs[numChilds-1];
}